package pboweaverplugin.builder;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.FileHandler;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.SerializationUtils;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IMarker;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceDelta;
import org.eclipse.core.resources.IResourceDeltaVisitor;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.OperationCanceledException;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.QualifiedName;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.console.ConsolePlugin;
import org.eclipse.ui.console.IConsole;
import org.eclipse.ui.console.IConsoleConstants;
import org.eclipse.ui.console.IConsoleManager;
import org.eclipse.ui.console.IConsoleView;
import org.eclipse.ui.console.MessageConsole;
import org.eclipse.ui.console.MessageConsoleStream;

import pboweaverplugin.properties.WeaverPropertyPage;

public class WeaverBuilder extends IncrementalProjectBuilder {
	
	List<IFile> sources = new ArrayList<IFile>();
	
	Logger log = Logger.getLogger(this.getClass().getCanonicalName());
	
	Map<String, String> extensionToWovenExtension = null;
	
	public static Pattern unclosedChoice = Pattern.compile("ERROR: ##CHOICE (.+?) in file (.+?) on line (\\d+) must be closed");
	public static Pattern missingParamDeclaration = Pattern.compile("ERROR: parameter (.+?), used on line (.+?) of file (.+?) has not been defined");
	public static Pattern mismatchedEndChoice = Pattern.compile("ERROR: ##BEGIN CHOICE (.+?) on line (\\d+) does not match ##END CHOICE (.+?) on line (\\d+) of (.+)\\.");
	public static Pattern duplicateParameterDeclaration = Pattern.compile("ERROR: Duplicate param ##(.+?) found in (.+?) on line (\\d+).  First saw ##(.+?) in (.+?) on line (\\d+)\\.");
	public static Pattern uninitializedDefaultValue = Pattern.compile("ERROR: parameter ##(.+?) found in (.+?) on line (\\d+) must have a default value\\.");
	public static Pattern namedInstanceChoiceConflict = Pattern.compile("ERROR: choice ##(.+?) on line (\\d+) of source (.+?) conflicts with choice ##(.+?) with named instance '(.+?)' on line (\\d+) of source (.+?)\\.$");
	public static Pattern choiceConflict = Pattern.compile("ERROR: choice ##(.+?) with named instance '(.+?)' on line (\\d+) of source (.+?) conflicts with choice ##(.+?) on line (\\d+) of source (.+?)\\.$");
	public static Pattern errorPattern = Pattern.compile("ERROR: .+");

	private static final String CONSOLE_NAME = "PbOConsole";
	
	protected static final Pattern instPattern = Pattern.compile("([\\w]+=((\".*\")|('.*')|[\\S]+))");
	
	// These will be set in initializeWeaver, which is called in startupOnInitialize
	private File weaver = null;
	private File weaverFolder = null;
	
	//FIXME: hard coded to src directory
	// this also affects the Runtime.getRuntime().exec() function call, as I pass this
	// File as the current working directory.

	private File output_dir = null;
	
	//------------------------------------------------------------------------
	// This constructor also sets up the supported file extensions this
	// 	weaver builder supports.  Right now, that is only C/C++ projects.
	// 	However, the behaviour of the weaver is undetermined when a project
	// 	consists of both C and C++ enriched sources.  Once a better way 
	// 	of determining the nature of the project is found, so that this
	//	HashMap can be properly initialized, the undetermined behaviour
	// 	should go away.
	//------------------------------------------------------------------------
	public WeaverBuilder() {
		super();
		extensionToWovenExtension = new HashMap<String,String>();
		extensionToWovenExtension.put("pbo-c","c");
		extensionToWovenExtension.put("pbo-cpp","cpp");
	}
	
	@Override
	/*
	 * (non-Javadoc)
	 * allows the builder to initialize variables that depend on getting the project
	 * since the project will have been set up by now
	 * @see org.eclipse.core.resources.IncrementalProjectBuilder#startupOnInitialize()
	 */
	
	protected void startupOnInitialize() {
		super.startupOnInitialize();
		retrieveState();
		
		// set the class logger to output to a file, so that we can debug the plugin
		
		try {
			log.addHandler(new FileHandler(ResourcesPlugin.getPlugin().getStateLocation().toString()+File.separatorChar+getProject().getName()+File.separatorChar+"weaver.log"));
		} catch (SecurityException e) {
			log.throwing("WeaverBuilder", "startupOnInitialize", e);
			e.printStackTrace();
			throw new RuntimeException("Unable to register FileHandler for Logger: "+e.getMessage());
		} catch (IOException e) {
			log.throwing("WeaverBuilder","startupOnInitialize", e);
			e.printStackTrace();
			throw new RuntimeException("Unable to register FileHandler for Logger: "+e.getMessage());
		}
		
		initializeWeaver();
	}
	
	class SampleDeltaVisitor implements IResourceDeltaVisitor {

		public boolean visit(IResourceDelta delta) throws CoreException {
			IResource resource = delta.getResource();
			
			// it is ONLY in the case that a file has removed that we should do something differently,
			// otherwise, we check if we need to add the file to the list of sources or not
			if (delta.getKind() == IResourceDelta.REMOVED && resource instanceof IFile && !resource.isDerived()) {
				IFile file = (IFile)resource;
				
				// When we rename or remove pbo-c or pbo-cpp files from a project, the woven
				// hidden .c or .cpp files still linger, causing the Makefile generator tool within
				// Eclipse to get confused and start compiling extra sources
				//
				// Here, we remove these lingering .c and .cpp files so that the Makefile
				// tool no longer detects these extra old files
				if (extensionToWovenExtension.keySet().contains(file.getFileExtension())) {
					numFilesChanged++;
					String pboExtension = file.getFileExtension();
					String hiddenFile = file.getName().toString().replaceFirst(pboExtension,extensionToWovenExtension.get(pboExtension));
					String parentPath = file.getParent().getLocation().toString();
					File f = new File(parentPath, "."+hiddenFile);
					log.info("[WeaverBuilder::SampleDeltaVisitor::visit] Removing file "+f);
					if (f.exists()) {
						f.delete();
					}
				}
				
				if (sources.contains(file)) {
					sources.remove(file);
					return true;
				}
			}
			
			return checkSourceFiles(resource);
		}
	}
	
	//------------------------------------------------------------------------
	// Here I assume that the Sample Delta Visitor will handle all file
	// deletions for me.  It is unclear how we could detect a file deletion
	// with this type of ResourceVisitor since in the DeltaVisitor, we have
	// delta.getKind() to tell us.
	//
	// In the SampleDeltaVistor code, it is, indeed, the case that the 
	// SampleDeltaVisitor will detect and handle the deletion of files within
	// the project
	//------------------------------------------------------------------------

	class SampleResourceVisitor implements IResourceVisitor {
		public boolean visit(IResource resource) {
			return checkSourceFiles(resource);
		}
	}
	
	private boolean checkSourceFiles(IResource resource) {
		IFile file = null;
		
		if (resource instanceof IFile && !resource.isDerived()) {
			file = (IFile)resource;
			
			if (extensionToWovenExtension.keySet().contains(file.getFileExtension())) {
				numFilesChanged++;
				if (!sources.contains(file)) {
					log.info("Adding file "+file+" to list of sources");
					sources.add(file);
				}
			}
		}
		
		return true;
	}

	public static final String BUILDER_ID = "PbOWeaverPlugin.sampleBuilder";

	private static final String MARKER_TYPE = "PbOWeaverPlugin.xmlProblem";
	
	// In the future, we may consider moving this and deleteMarkers to an AbstractWeaver class
	// since it may be the case that, in the future, this will become the responsibility of the weaver
	// and not the builder

	private void addMarker(IFile file, String message, int lineNumber,
			int severity) {
		try {
			IMarker marker = file.createMarker(MARKER_TYPE);
			marker.setAttribute(IMarker.MESSAGE, message);
			marker.setAttribute(IMarker.SEVERITY, severity);
			if (lineNumber == -1) {
				lineNumber = 1;
			}
			marker.setAttribute(IMarker.LINE_NUMBER, lineNumber);
		} catch (CoreException e) {
		}
	}
	
	private void deleteMarkers(IFile file) {
		try {
			file.deleteMarkers(MARKER_TYPE, false, IResource.DEPTH_ZERO);
		} catch (CoreException ce) {
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.core.internal.events.InternalBuilder#build(int,
	 *      java.util.Map, org.eclipse.core.runtime.IProgressMonitor)
	 */
	
	private int numFilesChanged = 0;
	
	protected IProject[] build(int kind, Map args, IProgressMonitor monitor)
			throws CoreException {
		
		numFilesChanged = 0;
		
		if (kind == FULL_BUILD) {
			fullBuild(monitor);
		} else {
			IResourceDelta delta = getDelta(getProject());
			if (delta == null) {
				fullBuild(monitor);
			} else {
				incrementalBuild(delta, monitor);
			}
		}
		
		saveState();
		if (numFilesChanged > 0) {
			weaveFiles(monitor);
		}
		
		return null;
	}

	protected void fullBuild(final IProgressMonitor monitor) throws CoreException {
		try {
			getProject().accept(new SampleResourceVisitor());
		} catch (CoreException e) { }
	}

	protected void incrementalBuild(IResourceDelta delta,
			IProgressMonitor monitor) throws CoreException {
		// the visitor does the work.
		delta.accept(new SampleDeltaVisitor());
	}
	
	//------------------------------------------------------------------------
	// This method is only relevant for Java projects.
	//    * this method does nothing for C/C++ projects where eclipse uses the 
	//		Makefile's clean target
	//------------------------------------------------------------------------
	
	protected void clean(IProgressMonitor mon) throws CoreException {
		List<File> toDelete = new ArrayList<File>();
		
		String defaultPackage = "ca.ubc.beta.pbo".replaceAll(".", Character.valueOf(File.separatorChar).toString());
		
		for(IFile f : sources) {
			String pboExtension = f.getFileExtension();
			String filename = f.getLocation().toString().replaceFirst(pboExtension,extensionToWovenExtension.get(pboExtension));
			
			toDelete.add(new File(filename));
		}
		
		String projectDirectory = getProject().getLocation().toString();
		
		File supportFile = new File(projectDirectory+defaultPackage,"PBOSupport.java");
		
		toDelete.add(supportFile);
		
		for(File f : toDelete) {
			if (f.exists()) { f.delete(); }
		}
		
		refreshProject();
	}
	
	//------------------------------------------------------------------------
	// Refreshes the project's directory tree so that it can detect the 
	// new files that have been added as a result of the weaver
	//------------------------------------------------------------------------
	protected void refreshProject() throws CoreException {
		getProject().refreshLocal(IResource.DEPTH_INFINITE,null);
	}
	
	public String getResourcesLocation() {
		String directory = ResourcesPlugin.getPlugin().getStateLocation().toString()+File.separatorChar+getProject().getName();
		
		// If this directory does not exist (and most likely for a new project it will not)
		// then we should create it.  This is to fix the error that there will always be
		// a directory to write the sources.dat file.  If there were no directory to write
		// source.dat to, an error is not generated.
		
		File dir = new File(directory);
		if (!dir.exists()) dir.mkdirs();
		
		return directory;
	}
	
	protected void saveState() {
		try {
			ArrayList<String> files = new ArrayList<String>();
			
			for(IFile f : sources) {
				files.add(f.getProjectRelativePath().toString());
			}
			
			//String resourcesDirectory = ResourcesPlugin.getPlugin().getStateLocation().toString();
			
			File outData = new File(getResourcesLocation(),"sources.dat");
			
			log.info("Saving enriched sources list to "+outData);
			
			FileOutputStream fop = new FileOutputStream(outData);
			fop.write(SerializationUtils.serialize(files));
			fop.flush();
			fop.close();
			
		} catch (Exception ex) {
			log.severe(ex.getMessage());
		}
	}
	
	@SuppressWarnings("unchecked")
	protected void retrieveState() {
		File f = new File(getResourcesLocation(),"sources.dat");
		
		log.info("Retrieving state from "+f);
		
		if (f.exists()) {
			ArrayList<String> files = null;
			
			try {
				FileInputStream fins = new FileInputStream(f);
				files = (ArrayList<String>)SerializationUtils.deserialize(fins);
				
			} catch (FileNotFoundException ex) {
				log.severe(ex.getMessage());
			}
			
			if (files != null) {
				for (String s : files) {
					IPath path = new Path(s);
					
					log.info("Retrieving file "+path.toString()+" from sources.dat");
					
					sources.add(getProject().getFile(path));
				}
			}
		}
 	}
	
	private void initializeWeaver() {
		//---------------------------------- DEPRECATED ----------------------------------
		// this is old code that is used to get the file path to the weaver specific files
		//
		// since we are moving toward a plugin that embeds the weaver files,
		// this project property may no longer be necessary.
		//
		//String executable = getProject().getPersistentProperty(new QualifiedName("",WeaverPropertyPage.PATH_PROPERTY));
		
		//log.info("Executable from Preference Page: "+executable);
		
		//if (executable == null) {
		//	log.warning("No language specific weaver file specified.");
		//	//MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow(), "Missing Language Specific Weaver Path", "The path to the language specific weaver file is either incorrect or missing.  Please check your project properties to make sure this path is correct");
		//	throw new RuntimeException("The path to the language specific weaver file is either incorrect or missing.  Please check your project properties to make sure this path is correct.");
		//}
		
		//File weaver = new File(executable);
		//---------------------------------- DEPRECATED ----------------------------------
		
		// We get the folder entry (as opposed to the entry specific for the weaver file) so that
		// we can force the OSGi bundler to extract the entire folder (which includes things like module
		// folders that the weaver requires).
		try {
			URL url = Platform.getBundle("PbOWeaverPlugin").getEntry("/CWeaver");
			URI uri = FileLocator.toFileURL(url).toURI();
			
			this.weaverFolder = new File(uri);
			
			this.weaver = new File(weaverFolder,"pbo-weave-c.pl");
			
			log.info("Weaving with weaver at "+weaver.toString());
			
			//TODO: refactor this out!
			// for now, this is fine where it is, since this weaver builder is only 
			// designed to work for C/C++ projects.  However, in the future,
			// when, I think, we move to more of a C/C++ strategy design pattern,
			// using this Weaver Builder as a base, all this will be moved out
			// of Weaver builder and into the C/CPP strategy class.
			
			//File weaverFolder = weaver.getParentFile();
			File optc = new File(weaverFolder, "opt.c");
			File opth = new File(weaverFolder, "opt.h");
			
			this.output_dir = new File(getProject().getLocation().toString(), "src");
			
			File out_optc = new File(output_dir,"opt.c");
			File out_opth = new File(output_dir,"opt.h");
			
			// only do this if neither one of the opt files are already in src/
			if (!(out_optc.exists() || out_opth.exists())) {
				
				try {
					if (optc.exists()) { FileUtils.copyFile(optc, out_optc); }
					if (opth.exists()) { FileUtils.copyFile(opth, out_opth); }
				} catch (IOException e) { 
					log.severe("[WeaverBuilder::weaveFiles] IOException thrown when trying to copy opt.c and opt.h.  Reason: "+e.getLocalizedMessage());
					log.throwing("WeaverBuilder", "weaveFiles", e);
				}
			}
			
		} catch (IOException e) {
			log.throwing("WeaverBuilder", "initializeWeaver", e);
		} catch (URISyntaxException e) {
			log.throwing("WeaverBuilder", "initializeWeaver", e);
		}
	}
	
	protected void weaveFiles(IProgressMonitor monitor) throws CoreException {
		
		if (sources.size() <= 0) {
			log.severe("No source files to weave.");
			return;
		} else {
			for(IFile f : sources) {
				log.info("Trying to weave "+f);
			}
		}
		
		List<String> commands = new ArrayList<String>();
		
		String instantiations = getProject().getPersistentProperty(new QualifiedName("",WeaverPropertyPage.INST_PROPERTY));
		List<String> insts = parseInstantiations(instantiations);
		
		if (!(weaver.exists())) { 
			log.severe("Could not find file: "+weaver);
			//MessageDialog.openError(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), "Missing Language Specific Weaver Path", "The path to the language specific weaver file is either incorrect or missing.  Please check your project properties to make sure this path is correct");
			throw new RuntimeException("Could not find weaver executable with path: "+weaver+".  Please check that the weaver path specified is correct.");
		}
		
		commands.add("perl");// using the perl command means we don't need to set execute permissions
		
		commands.add(weaver.toString());
		
		//String output_dir = getProject().getLocation().toString()+File.separatorChar+"src";
		
		//File output_dir = new File(getProject().getLocation().toString(), "src");
		log.info("Weaver output directory: "+output_dir);
		
		commands.add("--output_dir="+output_dir);
		
		for(IFile file : sources) {
			String path = file.getLocation().toString();
			if (!commands.contains(path)) {
				commands.add(path);
			}
			
			deleteMarkers(file);
		}
		
		if (insts.size() > 0) {
			commands.add("--inst");
			commands.addAll(insts);
		}
		
		String[] execute = commands.toArray(new String[0]);
		
		log.info("Executing "+commands);
		
		checkCancel(monitor);
		
		// Display weaver output to a console page
		
		MessageConsole con = (MessageConsole) displayMyConsole();
		con.clearConsole();
		MessageConsoleStream out = con.newMessageStream();
		
		try { 
			
			checkCancel(monitor);
			
			String line;
			
			Process p = Runtime.getRuntime().exec(execute, null,output_dir); 
			BufferedReader brinput = new BufferedReader(new InputStreamReader(p.getInputStream()));
			BufferedReader brerror = new BufferedReader(new InputStreamReader(p.getErrorStream()));
			
			// Instead of having each error regular expression try and match against every line
			// of the weaver output, we filter out the weaver output looking for the special 
			// error string "Error: ".  We collect these error strings, so that each error
			// regex only has to match against those lines that truly contain errors
			//
			// This should lead to a reduction of total number of regular expression matches.
			
			List<String> weaverErrors = new ArrayList<String>();
			
			while((line=brinput.readLine())!=null) {
				
				out.println(line);
				
				checkCancel(monitor);
				
				Matcher matcher = errorPattern.matcher(line);
				if (matcher.find()) { weaverErrors.add(line); }
				
			}
			brinput.close();
			
			int numErrors = 0;
			
			numErrors += checkUnclosedChoiceError(weaverErrors,sources).size();
			numErrors += checkMismatchedEndChoiceDeclarationError(weaverErrors,sources).size();
			numErrors += checkMissingParameterDeclarationError(weaverErrors,sources).size();
			numErrors += checkDuplicateParameterDeclarationError(weaverErrors,sources).size();
			numErrors += checkUninitializedDefaultValueError(weaverErrors, sources).size();
			numErrors += checkChoiceConflictError(weaverErrors, sources).size();
			numErrors += checkInstanceNameChoiceConflictError(weaverErrors, sources).size();
			
			while((line=brerror.readLine())!=null) {
				checkCancel(monitor);
				
				out.println(line);
				
				log.severe(line);
			}
			brerror.close();
			p.waitFor();
			
			getProject().refreshLocal(IResource.DEPTH_INFINITE, monitor);
			
			if (numErrors > 0) {
				throw new CoreException(new IStatus() {

					@Override
					public IStatus[] getChildren() {
						return new IStatus[0];
					}

					@Override
					public int getCode() {
						return 0;
					}

					@Override
					public Throwable getException() {
						return new IllegalStateException("ERROR: Weaver returned syntax errors that must be corrected");
					}

					@Override
					public String getMessage() {
						return "Default values for some parameters were not supplied";
					}

					@Override
					public String getPlugin() {
						return ResourcesPlugin.getPlugin().toString();
					}

					@Override
					public int getSeverity() {
						return IStatus.ERROR;
					}

					@Override
					public boolean isMultiStatus() {
						return false;
					}

					@Override
					public boolean isOK() {
						return false;
					}

					@Override
					public boolean matches(int severityMask) {
						return (severityMask & IStatus.WARNING)>0;
					}
					
				});
			}
			
		} catch (IOException e) { 
			System.out.println("ERROR: Could not execute weaver: "+e.getMessage()); 
			e.printStackTrace(); 
		} catch (InterruptedException ex) {
			log.warning("ERROR: Weaving Process got interrupted: "+ex.getMessage());
			ex.printStackTrace();
		}
		
		refreshProject();
	}

	protected void checkCancel(IProgressMonitor monitor) {
		if (monitor.isCanceled()) {
			throw new OperationCanceledException();
		}
		
	}
	
	protected MessageConsole findConsole(String name) {
		ConsolePlugin plugin = ConsolePlugin.getDefault();
		IConsoleManager conMan = plugin.getConsoleManager();
		IConsole[] existing = conMan.getConsoles();
		
		for(int i=0; i<existing.length; i++) {
			if (name.equals(existing[i].getName())) {
				return (MessageConsole) existing[i];
			}
		}
		
		MessageConsole myConsole = new MessageConsole(name,null);
		conMan.addConsoles(new IConsole[] {myConsole});
		
		return myConsole;
		
	}
	
	protected IConsole displayMyConsole() {
		final IConsole myConsole = findConsole(CONSOLE_NAME);
		
		// in order to switch the view to the console
		// we must wrap it around a Display.asyncExec so that
		// getWorkbench().getActiveWorkbenchWindow() does not return
		// null
		
		// since switching to the console view is a relatively
		// fast operation, the UI should not become unresponsive
		// during this time.
		
		Display.getDefault().asyncExec(new Runnable() {
			@Override
			public void run() {
				IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
				String id = IConsoleConstants.ID_CONSOLE_VIEW;
				try { 
					IConsoleView view = (IConsoleView)page.showView(id); 
					view.display(myConsole);
				} catch (PartInitException ex) {
					log.severe(ex.getMessage());
				}	
			}
		});
		
		return myConsole;	
	}
	
	//------------------------------------------------------------------------
	// Parses the input from the property page and extracts all instantiations
	// so that we can pass them as command line arguments to the weaver
	//------------------------------------------------------------------------
	
	protected List<String> parseInstantiations(String insts) {
		log.info("Parsing instantiations: "+insts);
		
		List<String> ret = new ArrayList<String>();
		
		if (insts == null) { return ret; }
		
		Matcher match = instPattern.matcher(insts);
		
		while(match.find()) {
			log.info("Found instantiation: "+match.group(1));
			
			ret.add("--"+match.group(1));
		}
		
		return ret;
	}
	
	//------------------------------------------------------------------------
	// Checks to make sure that the output from the Weaver does not
	// contain the string denoting the weaver detecting that an 
	// unclosed choice existed in the woven sources
	//
	// returns a set of IFiles that contain the error or null if no such files
	// exists in the project or there is no match on the lines
	//------------------------------------------------------------------------
	
	public Set<IFile> checkUnclosedChoiceError(List<String> lines, List<IFile> sources) {
		Set<IFile> ret = new HashSet<IFile>();
		
		for(String line : lines) {
			Matcher unclosedChoiceMatcher = unclosedChoice.matcher(line);
			if (unclosedChoiceMatcher.find()) {
				String choice 	= unclosedChoiceMatcher.group(1);
				String file 	= unclosedChoiceMatcher.group(2);
				String lineNo	= unclosedChoiceMatcher.group(3);
				
				log.info("Found Error: [CHOICE:"+choice+"][FILE:"+file+"][LINE:"+lineNo+"]");
				
				for(IFile f : sources) {
					if (f.getLocation().toString().equals(file)) {
						addMarker(f,"Unclosed Choice: "+choice,Integer.parseInt(lineNo),IMarker.SEVERITY_ERROR);
						ret.add(f);
					}
				}
			}
		}
		
		return ret;
	}
	
	//------------------------------------------------------------------------
	// Checks to make sure a line of output from the Weaver does not
	// contain the string denoting the weaver detecting that a parameter
	// has been used that has not first been declared in any enriched source
	//
	// returns the IFile that contained the error or null if no such file
	// exists in the project or there is no match on the line
	//------------------------------------------------------------------------
	
	public Set<IFile> checkMissingParameterDeclarationError(List<String> lines, List<IFile> sources) {
		
		Set<IFile> ret = new HashSet<IFile>();
		
		for(String line : lines) {
			Matcher missingParamDeclarationMatcher = missingParamDeclaration.matcher(line);
			
			if (missingParamDeclarationMatcher.find()) {
				String param 	= missingParamDeclarationMatcher.group(1);
				String lineNo 	= missingParamDeclarationMatcher.group(2);
				String file 	= missingParamDeclarationMatcher.group(3);
				
				log.info("Found Error: [PARAM:"+param+"][FILE:"+file+"][LINE:"+lineNo+"]");
				
				for(IFile f: sources) {
					if (f.getLocation().toString().equals(file)) {
						addMarker(f,"Undeclared Parameter: "+param, Integer.parseInt(lineNo),IMarker.SEVERITY_ERROR);
						ret.add(f);
					}
				}
			}
		}
		
		return ret;
	}
	
	//------------------------------------------------------------------------
	// Checks to make sure that the output from the Weaver does not
	// contain the string denoting the weaver detecting an end choice's
	// choice name not matching that of the choice that it is closing
	//
	// returns a set of IFiles that contain the error or null if no such files
	// exists in the project or there is no match on the lines
	//------------------------------------------------------------------------
	
	public Set<IFile> checkMismatchedEndChoiceDeclarationError(List<String> lines, List<IFile> sources) {
		Set<IFile> ret = new HashSet<IFile>();
		
		
		for(String line : lines) {
 			Matcher mismatchedEndChoiceMatcher = mismatchedEndChoice.matcher(line);
			
			if (mismatchedEndChoiceMatcher.find()) {
				String beginChoice = mismatchedEndChoiceMatcher.group(1);
				String beginChoiceLine = mismatchedEndChoiceMatcher.group(2);
				String endChoice = mismatchedEndChoiceMatcher.group(3);
				String endChoiceLine = mismatchedEndChoiceMatcher.group(4);
				String file = mismatchedEndChoiceMatcher.group(5);
				
				log.info("Found Error: " +
						"[BEGIN CHOICE:"+beginChoice+"]" +
						"[LINE: "+beginChoiceLine+"]" +
						"[END CHOICE: "+endChoice+"]"+
						"[LINE: "+endChoiceLine+"]"+
						"[FILE: "+file+"]");
				
				for(IFile f : sources) {
					if (f.getLocation().toString().equals(file)) {
						addMarker(f, "Mismatched END CHOICE: "+endChoice+" for CHOICE "+beginChoice, Integer.parseInt(endChoiceLine), IMarker.SEVERITY_ERROR);
						ret.add(f);
					}
				}
			}
		}
		
		return ret;
	}
	
	public Set<IFile> checkUninitializedDefaultValueError(List<String> lines, List<IFile> sources) {
		Set<IFile> ret = new HashSet<IFile>();
		
		for (String line : lines) {
			Matcher uninitializedValueMatcher = uninitializedDefaultValue.matcher(line);
			
			if (uninitializedValueMatcher.find()) {
				String param = uninitializedValueMatcher.group(1);
				String source = uninitializedValueMatcher.group(2);
				int lineNum = Integer.parseInt(uninitializedValueMatcher.group(3));
				
				log.info("Found uninitialized value: "+
						"[PARAM: "+param+"] "+
						"[FILE: " + source+"] "+
						"[LINE: " + lineNum+"]"
				);
				
				for(IFile f : sources) {
					if (f.getLocation().toString().equals(source)) {
						addMarker(f, "Default value not initialized for parameter ##"+param, lineNum, IMarker.SEVERITY_ERROR);
						ret.add(f);
					}
				}
			}
		}
		
		return ret;
	}
	
	//------------------------------------------------------------------------
	// Class Pair
	//
	// This is a support class used for duplicate parameter entries error 
	// checking.  Since we only want to add a marker for each 
	// <filename,line number, parameter> triplet, this class acts as the
	// composite from which we create the triplet.
	//------------------------------------------------------------------------
	private static class Pair<FIRST,SECOND> implements Comparable<Pair<FIRST,SECOND>> {
		public final FIRST first;
		public final SECOND second;
		
		private Pair(FIRST first, SECOND second) {
			this.first = first;
			this.second = second;
		}
		
		public static <FIRST,SECOND> Pair<FIRST,SECOND> of(FIRST first, SECOND second) {
			return new Pair<FIRST,SECOND>(first,second);
		}
		@Override
		public int compareTo(Pair<FIRST, SECOND> o) {
			int cmp = compare(first, o.first);
			return cmp == 0 ? compare(second, o.second) : cmp;
		}
		
		private int compare(Object o1, Object o2) {
			return o1 == null ? o2 == null ? 0 : -1 : o2 == null ? 1 : ((Comparable<Object>)o1).compareTo(o2);
		}
		
		public int hashCode() {
			return 31*hashcode(first)+hashcode(second);
		}
		
		private int hashcode(Object o) {
			return o == null ? 0 : o.hashCode();
		}
		
		public boolean equals(Object obj) {
			if (!(obj instanceof Pair)) return false;
			if (this == obj) return true;
			
			Pair<FIRST,SECOND> o = (Pair<FIRST,SECOND>)obj;
			return equal(first, o.first) && equal(second, o.second);
		}
		
		public boolean equal(Object o1, Object o2) {
			return o1 == null ? o2 == null : (o1 == o2 || o1.equals(o2));
		}
		
		public String toString() { return "("+first+","+second+")"; }
		
	}
	
	//------------------------------------------------------------------------
	// Class Triplet
	//
	// This is the class that we actually use to implement the 
	// <file, line number, param> triplet discussed in the above class
	//
	// We inherit from Pair so that we get the equals/compares methods for
	// free.  We also obfuscate the first and second class members by
	// our own first and second class members.  This does not seem to be
	// too big of a problem in terms of .equals().
	//------------------------------------------------------------------------
	
	private static class Triplet<FIRST, SECOND, THIRD> extends Pair<Pair<FIRST,SECOND>,THIRD> {
		public final FIRST first;
		public final SECOND second;
		public final THIRD third;
		
		private Triplet(FIRST first, SECOND second, THIRD third) {
			super(Pair.of(first,second), third);
			this.first = super.first.first;
			this.second = super.first.second;
			this.third = super.second;
		}
		
		public static <FIRST,SECOND,THIRD> Triplet<FIRST,SECOND,THIRD> of(FIRST first, SECOND second, THIRD third) {
			return new Triplet<FIRST,SECOND,THIRD>(first,second,third);
		}
	}
	
	//------------------------------------------------------------------------
	// Checks to make sure that the output from the Weaver does not
	// contain the string denoting the weaver detecting the declaration of
	// any PbO parameter multiple times
	//
	// returns a set of IFiles that contain the error or null if no such files
	// exists in the project or there is no match on the lines
	//------------------------------------------------------------------------
	
	public Set<IFile> checkDuplicateParameterDeclarationError(List<String> lines, List<IFile> sources) {
		
		Set<Triplet<IFile,String,String>> pairs = new HashSet<Triplet<IFile,String,String>>();
		
		for(String line : lines) {
			Matcher matcher = duplicateParameterDeclaration.matcher(line);
			if (matcher.find()) {
				String param = matcher.group(1);
				String file = matcher.group(2);
				String lineNumber = matcher.group(3);
				String origFile = matcher.group(5);
				String origLineNumber = matcher.group(6);
				
				log.info("Found Error: "+
					"[PARAM: "+param+"]"+
					"[LINE: "+lineNumber+"]"+
					"[FILE: "+file+"]"+
					"[ORIGINAL FILE: "+origFile+"]"+
					"[ORIGIANL LINE: "+origLineNumber+"]");
				
				for(IFile f : sources) {
					if (f.getLocation().toString().equals(file)) {
						pairs.add(Triplet.of(f,lineNumber,param));
					}
					if (f.getLocation().toString().equals(origFile)) {
						pairs.add(Triplet.of(f,origLineNumber,param));
					}
				}	
			}
		}
		
		for(Triplet<IFile,String, String> p : pairs) {
			//System.out.println(p.first+" "+p.second+" "+p.third);
			
			addMarker(p.first, "Duplicate ##PARAM declaration for parameter "+p.third+".", Integer.parseInt(p.second), IMarker.SEVERITY_ERROR);
		}
		
		Set<IFile> ret = new HashSet<IFile>();
		
		for(Triplet<IFile,String,String> p : pairs) {
			ret.add(p.first);
		}
		
		return ret;
		
	}
	
	public Set<IFile> checkInstanceNameChoiceConflictError(List<String> lines, List<IFile> sources) {
		Set<Triplet<IFile,String,String>> pairs = new HashSet<Triplet<IFile,String,String>>();
		Set<IFile> ret = new HashSet<IFile>();
		
		for(String line : lines) {
			Matcher matcher = namedInstanceChoiceConflict.matcher(line);
			if (matcher.find()) {
				String param = matcher.group(1);
				String lineNum = matcher.group(2);
				String file = matcher.group(3);
				
				String conflictParam = matcher.group(4);
				String conflictLineNum = matcher.group(6);
				String conflictFile = matcher.group(7);
				
				log.info("[FOUND CONFLICT:][param: "+param+"]"
					+"[line: "+lineNum+"]"
					+"[source: "+file+"]"
					+"[conflicting param: "+conflictParam+"]"
					+"[conflicting line: "+conflictLineNum+"]"
					+"[conflicting source: "+conflictFile+"]"
						);
				
				for(IFile f : sources) {
					if (f.getLocation().toString().equals(file)) {
						pairs.add(Triplet.of(f, lineNum, param));
					}
					if (f.getLocation().toString().equals(conflictFile)) {
						pairs.add(Triplet.of(f, conflictLineNum, conflictParam));
					}
				}
				
			}
		}
		
		for(Triplet<IFile, String, String> p : pairs) {
			IFile file = p.first;
			int lineNum = Integer.parseInt(p.second);
			String param = p.third;
			
			addMarker(file, "Conflicting choice declaration for choice ##"+param, lineNum, IMarker.SEVERITY_ERROR);
			
			// also add the ifile to the returned list
			ret.add(p.first);
		}
		
		return ret;
		
	}
	
	public Set<IFile> checkChoiceConflictError(List<String> lines, List<IFile> sources) {
		Set<Triplet<IFile,String,String>> pairs = new HashSet<Triplet<IFile,String,String>>();
		Set<IFile> ret = new HashSet<IFile>();
		
		for(String line : lines) {
			Matcher matcher = choiceConflict.matcher(line);
			if (matcher.find()) {
				String param = matcher.group(1);
				String lineNum = matcher.group(3);
				String file = matcher.group(4);
				
				String conflictParam = matcher.group(5);
				String conflictLineNum = matcher.group(6);
				String conflictFile = matcher.group(7);
				
				log.info("[FOUND CONFLICT:][param: "+param+"]"
					+"[line: "+lineNum+"]"
					+"[source: "+file+"]"
					+"[conflicting param: "+conflictParam+"]"
					+"[conflicting line: "+conflictLineNum+"]"
					+"[conflicting source: "+conflictFile+"]"
						);
				
				for(IFile f : sources) {
					if (f.getLocation().toString().equals(file)) {
						pairs.add(Triplet.of(f, lineNum, param));
					}
					if (f.getLocation().toString().equals(conflictFile)) {
						pairs.add(Triplet.of(f, conflictLineNum, conflictParam));
					}
				}
				
			}
		}
		
		for(Triplet<IFile, String, String> p : pairs) {
			IFile file = p.first;
			int lineNum = Integer.parseInt(p.second);
			String param = p.third;
			
			addMarker(file, "Conflicting choice declaration for choice ##"+param, lineNum, IMarker.SEVERITY_ERROR);
			
			// also add the ifile to the returned list
			ret.add(p.first);
		}
		
		return ret;	
	}
}
